#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <netinet/in.h>
#include <netinet/tcp.h>

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>

#include "network.h"
#include "fdevent.h"
#include "log.h"
#include "file_cache.h"

#if defined __linux__ && defined HAVE_SENDFILE && defined HAVE_SYS_SENDFILE_H 
# define USE_LINUX_SENDFILE
#elif defined __FreeBSD__ && defined HAVE_SENDFILE && defined HAVE_SYS_UIO_H
# define USE_FREEBSD_SENDFILE
#elif defined __solaris__ && defined HAVE_SENDFILEV && defined HAVE_SYS_SENDFILE_H
# define USE_SOLARIS_SENDFILEV
#elif defined HAVE_MMAP
# define USE_MMAP
#endif

#if defined USE_LINUX_SENDFILE || defined USE_SOLARIS_SENDFILEV
/* sendfile() linux, sendfilev() solaris */
# include <sys/sendfile.h>
#elif defined USE_FREEBSD_SENDFILE
/* sendfile() freebsd */
# include <sys/uio.h>
#elif defined HAVE_MMAP
/* mmap() + write() */
#include <sys/mman.h>
#endif

#ifdef HAVE_OPENSSL
# include <openssl/ssl.h> 
# include <openssl/err.h> 
#endif

/*
 * This was 'borrowed' from tcpdump.
 *
 *
 * This is fun.
 *
 * In older BSD systems, socket addresses were fixed-length, and
 * "sizeof (struct sockaddr)" gave the size of the structure.
 * All addresses fit within a "struct sockaddr".
 *
 * In newer BSD systems, the socket address is variable-length, and
 * there's an "sa_len" field giving the length of the structure;
 * this allows socket addresses to be longer than 2 bytes of family
 * and 14 bytes of data.
 *
 * Some commercial UNIXes use the old BSD scheme, some use the RFC 2553
 * variant of the old BSD scheme (with "struct sockaddr_storage" rather
 * than "struct sockaddr"), and some use the new BSD scheme.
 *
 * Some versions of GNU libc use neither scheme, but has an "SA_LEN()"
 * macro that determines the size based on the address family.  Other
 * versions don't have "SA_LEN()" (as it was in drafts of RFC 2553
 * but not in the final version).  On the latter systems, we explicitly
 * check the AF_ type to determine the length; we assume that on
 * all those systems we have "struct sockaddr_storage".
 */
#ifndef SA_LEN
# ifdef HAVE_SOCKADDR_SA_LEN
#  define SA_LEN(addr)   ((addr)->sa_len)
# else /* HAVE_SOCKADDR_SA_LEN */
#  ifdef HAVE_STRUCT_SOCKADDR_STORAGE
static size_t get_sa_len(const struct sockaddr *addr) {
	switch (addr->sa_family) {
		
#   ifdef AF_INET
	case AF_INET:
		return (sizeof (struct sockaddr_in));
#   endif
		
#   ifdef AF_INET6
	case AF_INET6:
		return (sizeof (struct sockaddr_in6));
#   endif
		
	default:
		return (sizeof (struct sockaddr));
		
	}
}
#   define SA_LEN(addr)   (get_sa_len(addr))
#  else /* HAVE_SOCKADDR_STORAGE */
#   define SA_LEN(addr)   (sizeof (struct sockaddr))
#  endif /* HAVE_SOCKADDR_STORAGE */
# endif /* HAVE_SOCKADDR_SA_LEN */
#endif /* SA_LEN */



int network_server_init(server *srv, const char *host, unsigned short port) {
	int val;
	socklen_t addr_len;
	char hbuf[NI_MAXHOST];
	struct hostent *he;
	struct rlimit rlim;
	
#ifdef SO_ACCEPTFILTER
	struct accept_filter_arg afa;
#endif
	
#ifdef HAVE_IPV6
	if (srv->srvconf.use_ipv6) srv->srv_addr.plain.sa_family = AF_INET6;
	else srv->srv_addr.plain.sa_family = AF_INET;
	
	if (-1 == (srv->fd = socket(srv->srv_addr.plain.sa_family, SOCK_STREAM, IPPROTO_TCP))) {
		
		/* fallback to IPv4 */
		srv->srv_addr.plain.sa_family = AF_INET;
		if (-1 == (srv->fd = socket(srv->srv_addr.plain.sa_family, SOCK_STREAM, IPPROTO_TCP))) {
			log_error_write(srv, __FILE__, __LINE__, "ss", "socket failed:", strerror(errno));
			return -1;
		}
	}
#else
	srv->srv_addr.plain.sa_family = AF_INET;
	if (-1 == (srv->fd = socket(srv->srv_addr.plain.sa_family, SOCK_STREAM, IPPROTO_TCP))) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "socket failed:", strerror(errno));
		return -1;
	}
#endif
	val = 1;
	if (setsockopt(srv->fd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof val) < 0) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "socketsockopt failed:", strerror(errno));
		return -1;
	}
	
	/* set rlimits */
	if (0 != getrlimit(RLIMIT_NOFILE, &rlim)) {
		log_error_write(srv, __FILE__, __LINE__,
				"ss", "couldn't get 'max filedescriptors'",
				strerror(errno));
		return -1;
	}
	
	if (NULL == (srv->ev = fdevent_init(rlim.rlim_cur, srv->event_handler))) {
		log_error_write(srv, __FILE__, __LINE__,
				"s", "fdevent_init failed");
		return -1;
	}
	
	if (-1 == fdevent_fcntl_set(srv->ev, srv->fd)) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "fcntl failed:", strerror(errno));
		return -1;
	}
	
	switch(srv->srv_addr.plain.sa_family) {
#ifdef HAVE_IPV6
	case AF_INET6:
		memset(&srv->srv_addr, 0, sizeof(struct sockaddr_in6));
		srv->srv_addr.ipv6.sin6_family = AF_INET6;
		if (host == NULL) {
			srv->srv_addr.ipv6.sin6_addr = in6addr_any;
		} else {
			struct addrinfo hints, *res;
			int r;
			
			memset(&hints, 0, sizeof(hints));
			
			hints.ai_family   = AF_INET6;
			hints.ai_socktype = SOCK_STREAM;
			hints.ai_protocol = IPPROTO_TCP;
			
			if (0 != (r = getaddrinfo(host, NULL, &hints, &res))) {
				log_error_write(srv, __FILE__, __LINE__, 
						"sssss", "getaddrinfo failed: ", 
						gai_strerror(r), "'", host, "'");
				
				return -1;
			}
			
			memcpy(&(srv->srv_addr), res->ai_addr, res->ai_addrlen);
			
			freeaddrinfo(res);
		}
		srv->srv_addr.ipv6.sin6_port = htons(port);
		addr_len = sizeof(struct sockaddr_in6);
		break;
#endif
	case AF_INET:
		memset(&srv->srv_addr, 0, sizeof(struct sockaddr_in));
		srv->srv_addr.ipv4.sin_family = AF_INET;
		if (host == NULL) {
			srv->srv_addr.ipv4.sin_addr.s_addr = htonl(INADDR_ANY);
		} else {
			if (NULL == (he = gethostbyname(host))) {
				log_error_write(srv, __FILE__, __LINE__, 
						"sss", "gethostbyname failed: ", 
						hstrerror(h_errno), host);
				return -1;
			}
			
			if (he->h_addrtype != AF_INET) {
				log_error_write(srv, __FILE__, __LINE__, "sd", "addr-type != AF_INET: ", he->h_addrtype);
				return -1;
			}
			
			if (he->h_length != sizeof(struct in_addr)) {
				log_error_write(srv, __FILE__, __LINE__, "sd", "addr-length != sizeof(in_addr): ", he->h_length);
				return -1;
			}
			
			memcpy(&(srv->srv_addr.ipv4.sin_addr.s_addr), he->h_addr_list[0], he->h_length);
		}
		srv->srv_addr.ipv4.sin_port = htons(port);
		
		addr_len = sizeof(struct sockaddr_in);
		
		break;
	default:
		addr_len = 0;
		
		return -1;
	}
	
	if (0 != bind(srv->fd, (struct sockaddr *) &(srv->srv_addr), addr_len)) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "can't bind: ", strerror(errno));
		return -1;
	}
	
	if (-1 == listen(srv->fd, 128 * 8)) {
		log_error_write(srv, __FILE__, __LINE__, "ss", "listen failed: ", strerror(errno));
		return -1;
	}
#ifdef SO_ACCEPTFILTER
	/*
	 * FreeBSD accf_http filter
	 *
	 */
	bzero(&afa, sizeof(afa));
	strcpy(afa.af_name, "httpready");
	if (setsockopt(srv->fd, SOL_SOCKET, SO_ACCEPTFILTER, &afa, sizeof(afa)) < 0) {
		log_error_write(srv, __FILE__, __LINE__, "ssd", "(notice) can't set accept-filter 'httpready': ", strerror(errno), errno);
	}
#endif

	/* Lookup name: first get default local hostname */
	if (0 != gethostname(hbuf, sizeof(hbuf))) {
		log_error_write(srv, __FILE__, __LINE__,
				"ss", "gethostname failed:",
				strerror(errno));
		return -1;
		
	}
	buffer_copy_string(srv->hostname, hbuf);
	
	/* Lookup name: secondly try to get hostname for bind address */
	switch(srv->srv_addr.plain.sa_family) {
#ifdef HAVE_IPV6
	case AF_INET6:
		if (0 != getnameinfo((const struct sockaddr *)(&srv->srv_addr.ipv6), 
				     SA_LEN((const struct sockaddr *)&srv->srv_addr.ipv6), 
				     hbuf, sizeof(hbuf), NULL, 0, 0)) {
			log_error_write(srv, __FILE__, __LINE__,
					"SSSBS", "NOTICE: getnameinfo failed: ",
					hstrerror(h_errno), ", using '", srv->hostname, "' instead");
		} else {
			buffer_copy_string(srv->hostname, hbuf);
		}
		break;
#endif
	case AF_INET:
		if (NULL == (he = gethostbyaddr((char *)&srv->srv_addr.ipv4.sin_addr, sizeof(struct in_addr), AF_INET))) {
			log_error_write(srv, __FILE__, __LINE__,
					"SSSBS", "NOTICE: gethostbyaddr failed: ",
					hstrerror(h_errno), ", using '", srv->hostname, "' instead");
		} else {
			buffer_copy_string(srv->hostname, he->h_name);
		}
		break;
	}
	
	return 0;
}


int network_server_close(server *srv) {
	close(srv->fd);
	
	return 0;
}

int network_write_chunkset_plain(server *srv, connection *con, chunkset *c) {
	int ret = 0;
	int fd = con->fd;
	
	
#ifdef TCP_CORK	
	int i;

	i = 1;
	setsockopt(fd, IPPROTO_TCP, TCP_CORK, &i, sizeof(i));
#endif
	
	while(c->used) {
		int chunk_finished = 0;
		
#if 0
		log_error_write(srv, __FILE__, __LINE__, "sd", "c->used: ", c->used);
#endif
		switch(c->chunk[0]->type) {
		case MEM_CHUNK: {
			char * offset;
			size_t toSend;
			ssize_t r;
			
			offset = c->chunk[0]->b->ptr + c->chunk[0]->written;
			toSend = c->chunk[0]->b->used - 1 - c->chunk[0]->written;
			
#if 0
			log_error_write(srv, __FILE__, __LINE__, "dds", fd, c->chunk[0]->b->used - 1, offset);
#endif
			
			if ((r = write(fd, offset, toSend)) < 0) {
				log_error_write(srv, __FILE__, __LINE__, "ssd", "write failed: ", strerror(errno), fd);
				
				ret = -1;
				goto end;
			}
#if 0
			log_error_write(srv, __FILE__, __LINE__, "d", r);
#endif			
			c->chunk[0]->written += r;
			
			if (c->chunk[0]->written == c->chunk[0]->b->used - 1) {
				chunk_finished = 1;
			}
			
			break;
		}
		case FILE_CHUNK: {
#ifdef USE_MMAP
			char *p = NULL;
#endif
#if defined USE_FREEBSD_SENDFILE
			off_t r;
#else
			ssize_t r;
#endif
			off_t offset;
			size_t toSend;
			
			if (NULL == (con->fce = file_cache_get_entry(srv, con, c->chunk[0]->b, con->fce))) {
				log_error_write(srv, __FILE__, __LINE__, "sb",
						strerror(errno), c->chunk[0]->b);
				ret = -1;
				goto end;
			}
			
			offset = c->chunk[0]->offset + c->chunk[0]->written;
			toSend = c->chunk[0]->length - c->chunk[0]->written;
			
#if defined USE_LINUX_SENDFILE
			/* Linux sendfile() */
			if (-1 == (r = sendfile(fd, con->fce->fd, &offset, toSend))) {
				if (errno != EAGAIN) {
					log_error_write(srv, __FILE__, __LINE__, "ssd", "sendfile: ", strerror(errno), errno);
					
					ret = -1;
					goto end;
				}
				
				r = 0;
			}
			
			c->chunk[0]->written += r;
#elif defined USE_FREEBSD_SENDFILE
			/* FreeBSD sendfile() */

			if (-1 == sendfile(con->fce->fd, fd, offset, toSend, NULL, &r, 0)) {
				if (errno != EAGAIN) {
					log_error_write(srv, __FILE__, __LINE__, "ssd", "sendfile: ", strerror(errno), errno);
					
					ret = -1;
					goto end;
				}
			}

			c->chunk[0]->written += r;
#elif defined USE_MMAP
			if (MAP_FAILED == (p = mmap(0, con->fce->st.st_size, PROT_READ, MAP_SHARED, con->fce->fd, 0))) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "mmap failed: ", strerror(errno));
				
				ret = -1;
				goto end;
			}
			
			
			if ((r = write(fd, p + offset, toSend)) <= 0) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "write failed: ", strerror(errno));
				
				ret = -1;
				goto end;
			}

			c->chunk[0]->written += r;
			
			munmap(p, c->chunk[0]->length);
#else
			buffer_prepare_copy(srv->tmp_buf, toSend);
			
			lseek(con->fce->fd, offset, SEEK_SET);
			if (-1 == (toSend = read(con->fce->fd, srv->tmp_buf->ptr, toSend))) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "read: ", strerror(errno));
				
				ret = -1;
				goto end;
			}
			
			if (-1 == (r = write(fd, srv->tmp_buf->ptr, toSend))) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "write: ", strerror(errno));
				
				ret = -1;
				goto end;
			}
			
			c->chunk[0]->written += r;
#endif			
			
			if (c->chunk[0]->written == c->chunk[0]->length) {
				chunk_finished = 1;
			}
			
			break;
		}
		default:
			
			log_error_write(srv, __FILE__, __LINE__, "ds", c->used, "type not known");
			
			ret = -1;
			goto end;
		}
		
		if (chunk_finished) {
			/* remove chunk */
			if (c->used > 1) {
				size_t j;
				chunk *d;
				
				d = c->chunk[0];
				for (j = 1; j < c->used; j++) {
					c->chunk[j-1] = c->chunk[j];
				}
				c->chunk[c->used - 1] = d;
			}
			c->used--;
		} else {
			/* not finished yet */
			con->is_writable = 0;
			ret = 1;
			goto end;
		}
	}
	
end:

#ifdef TCP_CORK
	i = 0;
	setsockopt(fd, IPPROTO_TCP, TCP_CORK, &i, sizeof(i));
#endif
	
	return ret;
}

#ifdef HAVE_OPENSSL 

int network_write_chunkset_ssl(server *srv, connection *con, chunkset *c) {
	int ret = 0, ssl_r;
	int fd = con->fd;
	
#ifdef TCP_CORK	
	int i;

	i = 1;
	setsockopt(fd, IPPROTO_TCP, TCP_CORK, &i, sizeof(i));
#endif
	
	while(c->used) {
		int chunk_finished = 0, r;
		
		switch(c->chunk[0]->type) {
		case MEM_CHUNK:
			if ((r = SSL_write(con->ssl, c->chunk[0]->b->ptr, c->chunk[0]->b->used - 1)) <= 0) {
				switch ((ssl_r = SSL_get_error(con->ssl, r))) {
				case SSL_ERROR_WANT_WRITE:
					r = 0;
					break;
				default:
					log_error_write(srv, __FILE__, __LINE__, "sdds", "SSL:", 
							ssl_r, r,
							ERR_error_string(ERR_get_error(), NULL));
					
					ret = -1;
					goto end;
					
					break;
				}
			}
			
			c->chunk[0]->written += r;
			
			if (c->chunk[0]->written == c->chunk[0]->b->used - 1) {
				chunk_finished = 1;
			}
			
			break;
		case FILE_CHUNK: {
			int ifd;
			char *p = NULL, *s;
			ssize_t r;
			struct stat st;
			socklen_t optlen;
			off_t offset;
			size_t toSend;
			int chunk_size;
			
			stat(c->chunk[0]->b->ptr, &st);
			
			if (-1 == (ifd = open(c->chunk[0]->b->ptr, O_RDONLY))) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "open failed: ", strerror(errno));
				
				ret = -1;
				goto end;
			}
			
			/* should be done only once */
			optlen = sizeof(chunk_size);
			if (-1 == getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (void *) &chunk_size, &optlen)) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "getsockopt failed: ", strerror(errno));
				ret = -1;
				goto end;
			}
			
			offset = c->chunk[0]->offset + c->chunk[0]->written;
			toSend = c->chunk[0]->length - c->chunk[0]->written > chunk_size ? 
				chunk_size : c->chunk[0]->length - c->chunk[0]->written;
			
			
#if defined HAVE_MMAP
			if (MAP_FAILED == (p = mmap(0, st.st_size, PROT_READ, MAP_SHARED, ifd, 0))) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "mmap failed: ", strerror(errno));
				
				close(ifd);
				ret = -1;
				goto end;
			}
			
			s = p + offset;
#else
			buffer_prepare_copy(srv->tmp_buf, toSend);
			
			lseek(ifd, offset, SEEK_SET);
			if (-1 == (toSend = read(ifd, srv->tmp_buf->ptr, toSend))) {
				log_error_write(srv, __FILE__, __LINE__, "ss", "read failed: ", strerror(errno));
				
				close(ifd);
				
				ret = -1;
				goto end;
			}
			
			s = srv->tmp_buf->ptr;
#endif
			
			if ((r = SSL_write(con->ssl, s, toSend)) <= 0) {
				switch ((ssl_r = SSL_get_error(con->ssl, r))) {
				case SSL_ERROR_WANT_WRITE:
					r = 0;
					break;
				default:
					log_error_write(srv, __FILE__, __LINE__, "sdds", "SSL:", 
							ssl_r, r, 
							ERR_error_string(ERR_get_error(), NULL));
					
					close(ifd);
					ret = -1;
					goto end;
					
					break;
				}
			}
			
			
			
			c->chunk[0]->written += r;
			
#if defined HAVE_MMAP
			munmap(p, c->chunk[0]->length);
#endif
			
			close(ifd);
			
			if (c->chunk[0]->written == c->chunk[0]->length) {
				chunk_finished = 1;
			}
			
			break;
		}
		default:
			
			log_error_write(srv, __FILE__, __LINE__, "ds", c->used, "type not known");
			
			ret = -1;
			goto end;
		}
		
		if (chunk_finished) {
			/* remove chunk */
			size_t i;
			chunk *d;
			
			if (c->used > 1) {
				d = c->chunk[0];
				for (i = 1; i < c->used; i++) {
					c->chunk[i-1] = c->chunk[i];
				}
				c->chunk[c->used - 1] = d;
			}
			c->used--;
		} else {
			/* not finished yet */
			
			ret = 1;
			goto end;
		}
	}
	
end:

#ifdef TCP_CORK
	i = 0;
	setsockopt(fd, IPPROTO_TCP, TCP_CORK, &i, sizeof(i));
#endif
	
	return ret;
}
#endif


int network_write_chunkset(server *srv, connection *con, chunkset *c) {
#ifdef HAVE_OPENSSL
	if (srv->is_ssl) {
		return network_write_chunkset_ssl(srv, con, c);
	} else {
		return network_write_chunkset_plain(srv, con, c);
	}
#else
	return network_write_chunkset_plain(srv, con, c);
#endif
}
